@@ -149,6 +149,7 @@ _ffi.cdef('''
     #define MDB_REVERSEKEY ...
     #define MDB_TXN_FULL ...
     #define MDB_WRITEMAP ...
+    #define MDB_NOTLS ...
 
     // Helpers below inline MDB_vals. Avoids key alloc/dup on CPython, where
     // cffi will use PyString_AS_STRING when passed as an argument.
@@ -363,18 +364,37 @@ class Environment(object):
             documentation.
 
         `max_readers`:
-            Slots to allocate in lock file for read threads; attempts to open
-            the environment by more than this many clients simultaneously will
-            fail. Only meaningful for environments that aren't already open.
+            Maximum number of simultaneous read transactions. Can only be set
+            by the first process to open an environment, as it affects the size
+            of the lock file and shared memory area. Attempts to simultaneously
+            start more than this many *read* transactions will fail.
 
         `max_dbs`:
             Maximum number of databases available. If 0, assume environment
             will be used as a single database.
+
+        `initial_readers`:
+            Indicates the number of read transaction to preallocate in the
+            `reset` state. When `max_readers` is large and there are many open
+            read transactions, starting a read will incur a penalty due to the
+            linear scan required on the readers table. In this case read
+            transactions may be accelerated through preallocation.
+
+            Currently unused in both CPython and CFFI.
+
+        `spare_readers`:
+            Indicates the number of read transactions to maintain in the
+            `reset` state. When a read transaction completes and the internal
+            `spare_readers` list is smaller than this number, the transaction
+            is reset and appended to the list. Otherwise, the transaction is
+            aborted and its associated reader slot is freed.
+
+            Currently unused in both CPython and CFFI.
     """
     def __init__(self, path, map_size=10485760, subdir=True,
             readonly=False, metasync=True, sync=True, map_async=False,
-            mode=0644, create=True, writemap=False, max_readers=126,
-            max_dbs=0):
+            mode=0o644, create=True, writemap=False, max_readers=126,
+            max_dbs=0, initial_readers=0, spare_readers=0):
         envpp = _ffi.new('MDB_env **')
 
         rc = mdb_env_create(envpp)
@@ -398,7 +418,7 @@ class Environment(object):
         if create and subdir and not os.path.exists(path):
             os.mkdir(path)
 
-        flags = 0
+        flags = MDB_NOTLS
         if not subdir:
             flags |= MDB_NOSUBDIR
         if readonly:
@@ -613,6 +633,78 @@ class Environment(object):
         """Shortcut for :py:class:`lmdb.Transaction`"""
         return Transaction(self, **kwargs)
 
+    def get(self, key, default=None, db=None):
+        """Use a temporary read transaction to invoke
+        :py:meth:`Transaction.get`."""
+        with Transaction(self) as txn:
+            return txn.get(key, default, db)
+
+    def gets(self, keys, db=None):
+        """Use a temporary read transaction to invoke
+        :py:meth:`Transaction.get` for each key in `keys`. The returned value
+        is a dict containing one element for each key that existed."""
+        dct = {}
+        with Transaction(self) as txn:
+            for key in keys:
+                value = txn.get(key, None, db)
+                if value is not None:
+                    dct[key] = value
+        return dct
+
+    def put(self, key, value, dupdata=False, overwrite=True, append=False,
+            db=None):
+        """Use a temporary write transaction to invoke
+        :py:meth:`Transaction.put`."""
+        with Transaction(self, write=True) as txn:
+            txn.put(key, value, dupdata, overwrite, append, db)
+
+    def puts(self, items, dupdata=False, overwrite=True, append=False,
+             db=None):
+        """Use a temporary write transaction to invoke
+        :py:meth:`Transaction.put` as `put(x, y)` for each `(x, y)`
+        in `items`. Items must be a dict, or an iterable producing 2-tuples.
+        This function requires 2-tuples, no other sequence type is accepted.
+
+        Returns a list of :py:meth:`Transaction.put` return values.
+
+            .. code-block:: python
+
+                a_existed, b_existed = env.puts(overwrite=False, items={
+                    'a': '1',
+                    'b': '2'
+                })
+
+                if a_existed:
+                    print 'Did not overwrite a, it already existed.'
+                if b_existed:
+                    print 'Did not overwrite b, it already existed.'
+        """
+        if type(items) is dict:
+            items = items.iteritems()
+        with Transaction(self, write=True) as txn:
+            return [txn.put(key, value, dupdata, overwrite, append, db)
+                    for key, value in items]
+
+    def delete(self, key, value='', db=None):
+        """Use a temporary write transaction to invoke
+        :py:meth:`Transaction.delete`."""
+        with Transaction(self, write=True) as txn:
+            return txn.delete(key, value, db)
+
+    def deletes(self, keys, db=None):
+        """Use a temporary write transaction to invoke
+        :py:meth:`Transaction.delete` for each key in `keys`. Returns a list of
+        :py:meth:`Transaction.delete` return values."""
+        with Transaction(self, write=True) as txn:
+            return [txn.delete(key, '', db) for key in keys]
+
+    def cursor(self, buffers=False, db=None):
+        """Use a temporary read transaction to return a :py:class:`Cursor`. The
+        transaction will remain alive for as long as the cursor exists.
+        """
+        txn = Transaction(self, db, None, False, buffers)
+        return Cursor(db or self._db, txn)
+
 
 class _Database(object):
     """Internal database handle."""
@@ -646,16 +738,21 @@ open = Environment
 
 class Transaction(object):
     """
-    A transaction handle.
+    A transaction object. All operations require a transaction handle,
+    transactions may be read-only or read-write. Write transactions may not
+    span threads. Transaction objects implement the context manager protocol,
+    so that reliable release of the transaction happens even in the face of
+    unhandled exceptions:
 
-    All operations require a transaction handle, transactions may be read-only
-    or read-write. Transactions may not span threads; a transaction must only
-    be used by a single thread.
+        .. code-block:: python
 
-    Threads may only have a single transaction open for each environment.
+            # Transaction aborts correctly:
+            with env.begin(write=True) as txn:
+                crash()
 
-    Cursors may not span transactions; each cursor must be opened and closed
-    within a single transaction.
+            # Transaction commits automatically:
+            with env.begin(write=True) as txn:
+                txn.put('a', 'b')
 
     Equivalent to `mdb_txn_begin()
     <http://symas.com/mdb/doc/group__mdb.html#gad7ea55da06b77513609efebd44b26920>`_
@@ -696,11 +793,12 @@ class Transaction(object):
         self._val = _ffi.new('MDB_val *')
         self._to_py = _mvbuf if buffers else _mvstr
         self._deps = {}
-        flags = 0
         if write and env.readonly:
             raise Error('Cannot start write transaction with read-only env')
-        if not write:
-            flags |= MDB_RDONLY
+        if write:
+            flags = 0
+        else:
+            flags = MDB_RDONLY
         txnpp = _ffi.new('MDB_txn **')
         if parent:
             self._parent = parent
@@ -770,9 +868,9 @@ class Transaction(object):
                 raise Error("mdb_txn_abort", rc)
 
     def get(self, key, default=None, db=None):
-        """Fetch the first value matching `key`, otherwise return `default`. A
-        cursor must be used to fetch all values for a key in a `dupsort=True`
-        database.
+        """Fetch the first value matching `key`, returning `default` if `key`
+        does not exist. A cursor must be used to fetch all values for a key in
+        a `dupsort=True` database.
 
         Equivalent to `mdb_get()
         <http://symas.com/mdb/doc/group__mdb.html#ga8bf10cd91d3f3a83a34d04ce6b07992d>`_
@@ -1194,3 +1292,20 @@ class Cursor(object):
             raise Error("mdb_cursor_put", rc)
         self._cursor_get(MDB_GET_CURRENT)
         return True
+
+    def _iter_from(self, k, reverse):
+        """Helper for centidb. Please do not rely on this interface, it may be
+        removed in future.
+        """
+        if not k and not reverse:
+            found = self.first()
+        else:
+            found = self.set_range(k)
+        if reverse:
+            if not found:
+                self.last()
+            return self.iterprev()
+        else:
+            if not found:
+                return ()
+            return self.iternext()
