@@ -45,6 +45,7 @@
 
 #define NOINLINE __attribute__((noinline))
 
+
 static PyObject *Error;
 
 enum string_id {
@@ -56,6 +57,9 @@ enum string_id {
     DELETE_S,
     DUPDATA_S,
     DUPSORT_S,
+    FORCE_S,
+    ITEMS_S,
+    ITERITEMS_S,
     KEY_S,
     KEYS_S,
     MAP_ASYNC_S,
@@ -69,6 +73,7 @@ enum string_id {
     PARENT_S,
     PATH_S,
     READONLY_S,
+    REVERSE_S,
     REVERSE_KEY_S,
     SUBDIR_S,
     SYNC_S,
@@ -91,6 +96,9 @@ static const char *strings = (
     "delete\0"
     "dupdata\0"
     "dupsort\0"
+    "force\0"
+    "items\0"
+    "iteritems\0"
     "key\0"
     "keys\0"
     "map_async\0"
@@ -104,6 +112,7 @@ static const char *strings = (
     "parent\0"
     "path\0"
     "readonly\0"
+    "reverse\0"
     "reverse_key\0"
     "subdir\0"
     "sync\0"
@@ -126,6 +135,20 @@ extern PyTypeObject PyIterator_Type;
 
 struct EnvObject;
 
+#if PY_MAJOR_VERSION >= 3
+
+// Python 3.3 kindly exports the struct definitions for us.
+#   define MOD_RETURN(mod) return mod;
+#   define MODINIT_NAME PyInit_cpython
+#   define BUFFER_TYPE PyMemoryViewObject
+#   define MAKE_BUFFER() PyMemoryView_FromMemory("", 0, PyBUF_READ)
+#   define SET_BUFFER(buff, ptr, size) {\
+        (buff)->view.buf = (ptr); \
+        (buff)->view.len = (size); \
+        (buff)->hash = -1; \
+    }
+
+#else
 
 // So evil.
 typedef struct {
@@ -138,6 +161,23 @@ typedef struct {
     long b_hash;
 } PyBufferObject;
 
+#   define PyUnicode_InternFromString PyString_InternFromString
+#   define PyBytes_AS_STRING PyString_AS_STRING
+#   define PyBytes_GET_SIZE PyString_GET_SIZE
+#   define PyBytes_CheckExact PyString_CheckExact
+#   define PyBytes_FromStringAndSize PyString_FromStringAndSize
+#   define MOD_RETURN(mod) return
+#   define MODINIT_NAME initcpython
+#   define BUFFER_TYPE PyBufferObject
+#   define MAKE_BUFFER() PyBuffer_FromMemory("", 0)
+#   define SET_BUFFER(buf, ptr, size) {\
+        (buf)->b_hash = -1; \
+        (buf)->b_ptr = (ptr); \
+        (buf)->b_size = (size); \
+    }
+
+#endif
+
 struct list_head {
     struct lmdb_object *prev;
     struct lmdb_object *next;
@@ -217,7 +257,7 @@ static void invalidate(struct lmdb_object *parent)
     while(child) {
         struct lmdb_object *next = child->siblings.next;
         DEBUG("invalidating parent=%p child %p", parent, child)
-        child->ob_type->tp_clear((PyObject *) child);
+        Py_TYPE(child)->tp_clear((PyObject *) child);
         child = next;
     }
 }
@@ -248,7 +288,7 @@ typedef struct {
 
     MDB_txn *txn;
     int buffers;
-    PyBufferObject *key_buf;
+    BUFFER_TYPE *key_buf;
 } TransObject;
 
 typedef struct {
@@ -257,8 +297,8 @@ typedef struct {
 
     int positioned;
     MDB_cursor *curs;
-    PyBufferObject *key_buf;
-    PyBufferObject *val_buf;
+    BUFFER_TYPE *key_buf;
+    BUFFER_TYPE *val_buf;
     PyObject *item_tup;
     MDB_val key;
     MDB_val val;
@@ -336,20 +376,18 @@ dict_from_fields(void *o, const struct dict_field *fields)
 
 
 static PyObject * NOINLINE
-buffer_from_val(PyBufferObject **bufp, MDB_val *val)
+buffer_from_val(BUFFER_TYPE **bufp, MDB_val *val)
 {
-    PyBufferObject *buf = *bufp;
+    BUFFER_TYPE *buf = *bufp;
     if(! buf) {
-        buf = (PyBufferObject *) PyBuffer_FromMemory("", 0);
+        buf = (BUFFER_TYPE *) MAKE_BUFFER();
         if(! buf) {
             return NULL;
         }
         *bufp = buf;
     }
 
-    buf->b_hash = -1;
-    buf->b_ptr = val->mv_data;
-    buf->b_size = val->mv_size;
+    SET_BUFFER(buf, val->mv_data, val->mv_size);
     Py_INCREF(buf);
     return (PyObject *) buf;
 }
@@ -358,18 +396,30 @@ buffer_from_val(PyBufferObject **bufp, MDB_val *val)
 static PyObject *
 string_from_val(MDB_val *val)
 {
-    return PyString_FromStringAndSize(val->mv_data, val->mv_size);
+    return PyBytes_FromStringAndSize(val->mv_data, val->mv_size);
 }
 
 
 static int NOINLINE
 val_from_buffer(MDB_val *val, PyObject *buf)
 {
-    if(PyString_CheckExact(buf)) {
-        val->mv_data = PyString_AS_STRING(buf);
-        val->mv_size = Py_SIZE(buf);
+    if(PyBytes_CheckExact(buf)) {
+        val->mv_data = PyBytes_AS_STRING(buf);
+        val->mv_size = PyBytes_GET_SIZE(buf);
         return 0;
     }
+#if PY_MAJOR_VERSION >= 3
+    if(PyUnicode_CheckExact(buf)) {
+        char *data;
+        Py_ssize_t size;
+        if(! (data = PyUnicode_AsUTF8AndSize(buf, &size))) {
+            return -1;
+        }
+        val->mv_data = data;
+        val->mv_size = size;
+        return 0;
+    }
+#endif
     return PyObject_AsReadBuffer(buf,
         (const void **) &val->mv_data,
         (Py_ssize_t *) &val->mv_size);
@@ -448,7 +498,11 @@ parse_ulong(PyObject *obj, uint64_t *l, PyObject *max)
         type_error("Integer argument exceeds limit.");
         return -1;
     }
+#if PY_MAJOR_VERSION >= 3
+    *l = PyLong_AsUnsignedLongLongMask(obj);
+#else
     *l = PyInt_AsUnsignedLongLongMask(obj);
+#endif
     return 0;
 }
 
@@ -545,7 +599,7 @@ parse_args(int valid, int specsize, const struct argspec *argspec,
             if(val) {
                 if(set & (1 << i)) {
                     PyErr_Format(PyExc_TypeError, "duplicate argument: %s",
-                                 PyString_AS_STRING(kwd));
+                                 PyBytes_AS_STRING(kwd));
                     return -1;
                 }
                 if(parse_arg(spec, val, out)) {
@@ -564,6 +618,206 @@ parse_args(int valid, int specsize, const struct argspec *argspec,
 }
 
 
+// --------------------------------------------------------
+// Functionality shared between Transaction and Environment
+// --------------------------------------------------------
+
+
+static PyObject *
+generic_get(int valid, MDB_txn *txn, DbObject *db, int buffers,
+            BUFFER_TYPE **bptr, PyObject *args, PyObject *kwds)
+{
+    struct generic_get {
+        MDB_val key;
+        PyObject *default_;
+        DbObject *db;
+    } arg = {{0, 0}, Py_None, db};
+
+    static const struct argspec argspec[] = {
+        {ARG_BUF, KEY_S, OFFSET(generic_get, key)},
+        {ARG_OBJ, DEFAULT_S, OFFSET(generic_get, default_)},
+        {ARG_DB, DB_S, OFFSET(generic_get, db)}
+    };
+
+    if(parse_args(valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+
+    if(! arg.key.mv_data) {
+        return type_error("key must be given.");
+    }
+
+    MDB_val val;
+    int rc = mdb_get(txn, arg.db->dbi, &arg.key, &val);
+    if(rc) {
+        if(rc == MDB_NOTFOUND) {
+            Py_INCREF(arg.default_);
+            return arg.default_;
+        }
+        return err_set("mdb_get", rc);
+    }
+    if(buffers) {
+        return buffer_from_val(bptr, &val);
+    }
+    return string_from_val(&val);
+}
+
+static PyObject *
+generic_put(int valid, MDB_txn *txn, DbObject *db,
+            PyObject *args, PyObject *kwds)
+{
+    struct generic_put {
+        MDB_val key;
+        MDB_val value;
+        int dupdata;
+        int overwrite;
+        int append;
+        DbObject *db;
+    } arg = {{0, 0}, {0, 0}, 0, 1, 0, db};
+
+    static const struct argspec argspec[] = {
+        {ARG_BUF, KEY_S, OFFSET(generic_put, key)},
+        {ARG_BUF, VALUE_S, OFFSET(generic_put, value)},
+        {ARG_BOOL, DUPDATA_S, OFFSET(generic_put, dupdata)},
+        {ARG_BOOL, OVERWRITE_S, OFFSET(generic_put, overwrite)},
+        {ARG_BOOL, APPEND_S, OFFSET(generic_put, append)},
+        {ARG_DB, DB_S, OFFSET(generic_put, db)}
+    };
+
+    if(parse_args(valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+
+    int flags = 0;
+    if(! arg.dupdata) {
+        flags |= MDB_NODUPDATA;
+    }
+    if(! arg.overwrite) {
+        flags |= MDB_NOOVERWRITE;
+    }
+    if(arg.append) {
+        flags |= MDB_APPEND;
+    }
+
+    DEBUG("inserting '%.*s' (%d) -> '%.*s' (%d)",
+        (int)arg.key.mv_size, (char *)arg.key.mv_data,
+        (int)arg.key.mv_size,
+        (int)arg.value.mv_size, (char *)arg.value.mv_data,
+        (int)arg.value.mv_size)
+
+    int rc = mdb_put(txn, (arg.db)->dbi, &arg.key, &arg.value, flags);
+    if(rc) {
+        if(rc == MDB_KEYEXIST) {
+            Py_RETURN_FALSE;
+        }
+        return err_set("mdb_put", rc);
+    }
+    Py_RETURN_TRUE;
+}
+
+static PyObject *
+generic_delete(int valid, MDB_txn *txn, DbObject *db,
+               PyObject *args, PyObject *kwds)
+{
+    struct generic_delete {
+        MDB_val key;
+        MDB_val val;
+        DbObject *db;
+    } arg = {{0, 0}, {0, 0}, db};
+
+    static const struct argspec argspec[] = {
+        {ARG_BUF, KEY_S, OFFSET(generic_delete, key)},
+        {ARG_BUF, VALUE_S, OFFSET(generic_delete, val)},
+        {ARG_DB, DB_S, OFFSET(generic_delete, db)}
+    };
+
+    if(parse_args(valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+    MDB_val *val_ptr = arg.val.mv_size ? &arg.val : NULL;
+    int rc = mdb_del(txn, arg.db->dbi, &arg.key, val_ptr);
+    if(rc) {
+        if(rc == MDB_NOTFOUND) {
+             Py_RETURN_FALSE;
+        }
+        return err_set("mdb_del", rc);
+    }
+    Py_RETURN_TRUE;
+}
+
+static PyObject *
+make_trans(EnvObject *env, TransObject *parent, int write, int buffers)
+{
+    DEBUG("make_trans(env=%p, parent=%p, write=%d, buffers=%d)",
+        env, parent, write, buffers)
+    if(! env->valid) {
+        return err_invalid();
+    }
+
+    MDB_txn *parent_txn = NULL;
+    if(parent) {
+        if(! parent->valid) {
+            return err_invalid();
+        }
+        parent_txn = parent->txn;
+    }
+
+    if(write && env->readonly) {
+        return err_set("Cannot start write transaction with read-only env", 0);
+    }
+
+    TransObject *self = PyObject_New(TransObject, &PyTransaction_Type);
+    if(! self) {
+        return NULL;
+    }
+
+    int flags = (write && !env->readonly) ? 0 : MDB_RDONLY;
+    int rc = mdb_txn_begin(env->env, parent_txn, flags, &self->txn);
+    if(rc) {
+        PyObject_Del(self);
+        return err_set("mdb_txn_begin", rc);
+    }
+
+    OBJECT_INIT(self)
+    LINK_CHILD(env, self)
+    self->env = env;
+    Py_INCREF(env);
+    self->buffers = buffers;
+    self->key_buf = NULL;
+    return (PyObject *)self;
+}
+
+static PyObject *
+make_cursor(DbObject *db, TransObject *trans)
+{
+    if(! trans->valid) {
+        return err_invalid();
+    }
+    if(! db) {
+        db = trans->env->main_db;
+    }
+
+    CursorObject *self = PyObject_New(CursorObject, &PyCursor_Type);
+    int rc = mdb_cursor_open(trans->txn, db->dbi, &self->curs);
+    if(rc) {
+        PyObject_Del(self);
+        return err_set("mdb_cursor_open", rc);
+    }
+
+    DEBUG("sizeof cursor = %d", (int) sizeof *self)
+    OBJECT_INIT(self)
+    LINK_CHILD(trans, self)
+    self->positioned = 0;
+    self->key_buf = NULL;
+    self->val_buf = NULL;
+    self->key.mv_size = 0;
+    self->val.mv_size = 0;
+    self->item_tup = NULL;
+    self->trans = trans;
+    Py_INCREF(self->trans);
+    return (PyObject *) self;
+}
+
 // ----------------------------
 // Database
 // ----------------------------
@@ -655,7 +909,7 @@ PyTypeObject PyDatabase_Type = {
 static int
 env_clear(EnvObject *self)
 {
-    DEBUG("kiling env..")
+    DEBUG("killing env..")
     if(self->env) {
         INVALIDATE(self)
         DEBUG("Closing env")
@@ -759,7 +1013,7 @@ env_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
         }
     }
 
-    int flags = 0;
+    int flags = MDB_NOTLS;
     if(! arg.subdir) {
         flags |= MDB_NOSUBDIR;
     }
@@ -801,50 +1055,6 @@ fail:
     return NULL;
 }
 
-
-static PyObject *
-make_trans(EnvObject *env, TransObject *parent, int write, int buffers)
-{
-    DEBUG("make_trans(env=%p, parent=%p, write=%d, buffers=%d)",
-        env, parent, write, buffers)
-    if(! env->valid) {
-        return err_invalid();
-    }
-
-    MDB_txn *parent_txn = NULL;
-    if(parent) {
-        if(! parent->valid) {
-            return err_invalid();
-        }
-        parent_txn = parent->txn;
-    }
-
-    if(write && env->readonly) {
-        return err_set("Cannot start write transaction with read-only env", 0);
-    }
-
-    TransObject *self = PyObject_New(TransObject, &PyTransaction_Type);
-    if(! self) {
-        return NULL;
-    }
-
-    int flags = (write && !env->readonly) ? 0 : MDB_RDONLY;
-    int rc = mdb_txn_begin(env->env, parent_txn, flags, &self->txn);
-    if(rc) {
-        PyObject_Del(self);
-        return err_set("mdb_txn_begin", rc);
-    }
-
-    OBJECT_INIT(self)
-    LINK_CHILD(env, self)
-    self->env = env;
-    Py_INCREF(env);
-    self->buffers = buffers;
-    self->key_buf = NULL;
-    return (PyObject *)self;
-}
-
-
 static PyObject *
 env_begin(EnvObject *self, PyObject *args, PyObject *kwds)
 {
@@ -974,7 +1184,7 @@ env_path(EnvObject *self)
     if((rc = mdb_env_get_path(self->env, &path))) {
         return err_set("mdb_env_get_path", rc);
     }
-    return PyString_FromString(path);
+    return PyUnicode_FromString(path);
 }
 
 
@@ -1005,25 +1215,390 @@ env_stat(EnvObject *self)
 }
 
 static PyObject *
-env_sync(EnvObject *self, PyObject *arg)
+env_sync(EnvObject *self, PyObject *args)
+{
+    struct env_sync {
+        int force;
+    } arg = {0};
+
+    static const struct argspec argspec[] = {
+        {ARG_BOOL, FORCE_S, OFFSET(env_sync, force)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, NULL, &arg)) {
+        return NULL;
+    }
+
+    int rc = mdb_env_sync(self->env, arg.force);
+    if(rc) {
+        return err_set("mdb_env_sync", rc);
+    }
+    Py_RETURN_NONE;
+}
+
+static PyObject *
+env_get(EnvObject *self, PyObject *args, PyObject *kwds)
 {
     if(! self->valid) {
         return err_invalid();
     }
 
-    int force = arg == NULL;
-    if(arg) {
-        int force = PyObject_IsTrue(arg);
-        if(force == -1) {
-            return NULL;
+    MDB_txn *txn;
+    int rc = mdb_txn_begin(self->env, NULL, MDB_RDONLY, &txn);
+    if(rc) {
+        return err_set("mdb_txn_begin", rc);
+    }
+
+    PyObject *ret = generic_get(1, txn, self->main_db, 0, NULL, args, kwds);
+    mdb_txn_abort(txn);
+    return ret;
+}
+
+
+static PyObject *
+env_gets(EnvObject *self, PyObject *args, PyObject *kwds)
+{
+    struct env_gets {
+        PyObject *keys;
+        DbObject *db;
+    } arg = {NULL, self->main_db};
+
+    static const struct argspec argspec[] = {
+        {ARG_OBJ, KEYS_S, OFFSET(env_gets, keys)},
+        {ARG_DB, DB_S, OFFSET(env_gets, db)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+
+    if(! arg.keys) {
+        return type_error("keys must be given");
+    }
+
+    PyObject *iter = PyObject_GetIter(arg.keys);
+    if(! iter) {
+        return NULL;
+    }
+
+    PyObject *dict = PyDict_New();
+    if(! dict) {
+        Py_DECREF(iter);
+        return NULL;
+    }
+
+    MDB_txn *txn;
+    int rc = mdb_txn_begin(self->env, NULL, MDB_RDONLY, &txn);
+    if(rc) {
+        Py_DECREF(iter);
+        Py_DECREF(dict);
+        return err_set("mdb_txn_begin", rc);
+    }
+
+    PyObject *key_obj;
+    MDB_val key;
+    MDB_val val;
+
+    while((key_obj = PyIter_Next(iter)) != NULL) {
+        if(val_from_buffer(&key, key_obj)) {
+            break;
+        }
+
+        rc = mdb_get(txn, arg.db->dbi, &key, &val);
+        if(rc == 0) {
+            PyObject *val_obj = string_from_val(&val);
+            if(! val_obj) {
+                break;
+            }
+            rc = PyDict_SetItem(dict, key_obj, val_obj);
+            Py_DECREF(val_obj);
+            if(rc) {
+                break;
+            }
+        } else if(rc != MDB_NOTFOUND) {
+            err_set("mdb_get", rc);
+            break;
+        }
+        Py_DECREF(key_obj);
+    }
+
+    mdb_txn_abort(txn);
+    Py_DECREF(iter);
+    Py_XDECREF(key_obj);
+    if(PyErr_Occurred()) {
+        Py_CLEAR(dict);
+    }
+    return dict;
+}
+
+static PyObject *
+env_put(EnvObject *self, PyObject *args, PyObject *kwds)
+{
+    if(! self->valid) {
+        return err_invalid();
+    }
+
+    MDB_txn *txn;
+    int rc = mdb_txn_begin(self->env, NULL, 0, &txn);
+    if(rc) {
+        return err_set("mdb_txn_begin", rc);
+    }
+
+    PyObject *ret = generic_put(1, txn, self->main_db, args, kwds);
+    if(ret) {
+        if((rc = mdb_txn_commit(txn))) {
+            Py_DECREF(ret);
+            ret = err_set("mdb_txn_commit", rc);
+        }
+    } else {
+        mdb_txn_abort(txn);
+    }
+    return ret;
+}
+
+static PyObject *
+env_puts(EnvObject *self, PyObject *args, PyObject *kwds)
+{
+    struct env_puts {
+        PyObject *items;
+        int dupdata;
+        int overwrite;
+        int append;
+        DbObject *db;
+    } arg = {NULL, 0, 1, 0, self->main_db};
+
+    static const struct argspec argspec[] = {
+        {ARG_OBJ, ITEMS_S, OFFSET(env_puts, items)},
+        {ARG_BOOL, DUPDATA_S, OFFSET(env_puts, dupdata)},
+        {ARG_BOOL, OVERWRITE_S, OFFSET(env_puts, overwrite)},
+        {ARG_BOOL, APPEND_S, OFFSET(env_puts, append)},
+        {ARG_DB, DB_S, OFFSET(env_puts, db)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+
+    if(! arg.items) {
+        return type_error("items must be given");
+    }
+
+    PyObject *iter;
+    if(Py_TYPE(arg.items) == &PyDict_Type) {
+        iter = PyObject_CallMethodObjArgs(
+            arg.items, string_tbl[ITERITEMS_S], NULL);
+    } else {
+        iter = PyObject_GetIter(arg.items);
+    }
+    if(! iter) {
+        return NULL;
+    }
+
+    PyObject *list = PyList_New(0);
+    if(! list) {
+        Py_DECREF(iter);
+        return NULL;
+    }
+
+    MDB_txn *txn;
+    int rc = mdb_txn_begin(self->env, NULL, 0, &txn);
+    if(rc) {
+        Py_DECREF(iter);
+        Py_DECREF(list);
+        return err_set("mdb_txn_begin", rc);
+    }
+
+    int flags = 0;
+    if(! arg.dupdata) {
+        flags |= MDB_NODUPDATA;
+    }
+    if(! arg.overwrite) {
+        flags |= MDB_NOOVERWRITE;
+    }
+    if(arg.append) {
+        flags |= MDB_APPEND;
+    }
+
+    PyObject *item;
+    MDB_val key;
+    MDB_val val;
+
+    while((item = PyIter_Next(iter)) != NULL) {
+        if(! (PyTuple_Check(item) && PyTuple_GET_SIZE(item) == 2)) {
+            Py_DECREF(item);
+            type_error("puts() element type must be a 2-tuple.");
+            break;
+        }
+
+        if(val_from_buffer(&key, PyTuple_GET_ITEM(item, 0)) ||
+           val_from_buffer(&val, PyTuple_GET_ITEM(item, 1))) {
+            Py_DECREF(item);
+            break;
+        }
+
+        DEBUG("inserting '%.*s' (%d) -> '%.*s' (%d)",
+            (int)key.mv_size, (char *)key.mv_data, (int)key.mv_size,
+            (int)val.mv_size, (char *)val.mv_data, (int)val.mv_size)
+        rc = mdb_put(txn, arg.db->dbi, &key, &val, flags);
+        Py_DECREF(item);
+
+        PyObject *res;
+        if(rc == 0) {
+            res = Py_True;
+        } else if(rc == MDB_KEYEXIST) {
+            res = Py_False;
+        } else {
+            err_set("mdb_put", rc);
+            break;
+        }
+
+        if(PyList_Append(list, res)) {
+            break;
         }
     }
 
-    int rc = mdb_env_sync(self->env, force);
-    if(rc) {
-        return err_set("mdb_env_sync", rc);
+    DEBUG("got this far; list size now %d", (int) PyList_GET_SIZE(list))
+    Py_DECREF(iter);
+    if(PyErr_Occurred()) {
+        DEBUG("abort")
+        mdb_txn_abort(txn);
+        Py_CLEAR(list);
+    } else {
+        DEBUG("commit")
+        if((rc = mdb_txn_commit(txn))) {
+            err_set("mdb_txn_commit", rc);
+            Py_CLEAR(list);
+        }
     }
-    Py_RETURN_NONE;
+    return list;
+}
+
+static PyObject *
+env_delete(EnvObject *self, PyObject *args, PyObject *kwds)
+{
+    if(! self->valid) {
+        return err_invalid();
+    }
+
+    MDB_txn *txn;
+    int rc = mdb_txn_begin(self->env, NULL, 0, &txn);
+    if(rc) {
+        return err_set("mdb_txn_begin", rc);
+    }
+
+    PyObject *ret = generic_delete(1, txn, self->main_db, args, kwds);
+    if(ret) {
+        if((rc = mdb_txn_commit(txn))) {
+            Py_DECREF(ret);
+            ret = err_set("mdb_txn_commit", rc);
+        }
+    } else {
+        mdb_txn_abort(txn);
+    }
+    return ret;
+}
+
+static PyObject *
+env_deletes(EnvObject *self, PyObject *args, PyObject *kwds)
+{
+    struct env_deletes {
+        PyObject *keys;
+        DbObject *db;
+    } arg = {NULL, self->main_db};
+
+    static const struct argspec argspec[] = {
+        {ARG_OBJ, KEYS_S, OFFSET(env_deletes, keys)},
+        {ARG_DB, DB_S, OFFSET(env_deletes, db)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+
+    if(! arg.keys) {
+        return type_error("keys must be given");
+    }
+
+    PyObject *iter = PyObject_GetIter(arg.keys);
+    if(! iter) {
+        return NULL;
+    }
+
+    PyObject *list = PyList_New(0);
+    if(! list) {
+        Py_DECREF(iter);
+        return NULL;
+    }
+
+    MDB_txn *txn;
+    int rc = mdb_txn_begin(self->env, NULL, 0, &txn);
+    if(rc) {
+        return err_set("mdb_txn_begin", rc);
+    }
+
+    PyObject *key_obj;
+    MDB_val key;
+    while((key_obj = PyIter_Next(iter)) != NULL) {
+        if(val_from_buffer(&key, key_obj)) {
+            break;
+        }
+
+        rc = mdb_del(txn, arg.db->dbi, &key, NULL);
+        Py_DECREF(key_obj);
+
+        PyObject *res;
+        if(rc == 0) {
+            res = Py_True;
+        } else if(rc == MDB_NOTFOUND) {
+            res = Py_False;
+        } else {
+            err_set("mdb_del", rc);
+            break;
+        }
+        if(PyList_Append(list, res)) {
+            break;
+        }
+    }
+
+    if(PyErr_Occurred()) {
+        mdb_txn_abort(txn);
+        Py_CLEAR(list);
+    } else {
+        if((rc = mdb_txn_commit(txn))) {
+            Py_CLEAR(list);
+            err_set("mdb_txn_commit", rc);
+        }
+    }
+    return list;
+}
+
+static PyObject *
+env_cursor(EnvObject *self, PyObject *args, PyObject *kwds)
+{
+    // TODO: there is no benefit to this implementation. Once reader freelist
+    // is supported, should be possible to make meaningful optimizaiton here.
+    struct env_cursor {
+        int buffers;
+        DbObject *db;
+    } arg = { 0, self->main_db };
+
+    static const struct argspec argspec[] = {
+        {ARG_BOOL, BUFFERS_S, OFFSET(env_cursor, buffers)},
+        {ARG_DB, DB_S, OFFSET(env_cursor, db)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+
+    PyObject *trans = make_trans(self, NULL, 0, arg.buffers);
+    if(! trans) {
+        return NULL;
+    }
+
+    PyObject *cursor = make_cursor(arg.db, (TransObject *) trans);
+    Py_DECREF(trans);
+    return cursor;
 }
 
 static struct PyMethodDef env_methods[] = {
@@ -1034,7 +1609,14 @@ static struct PyMethodDef env_methods[] = {
     {"open_db", (PyCFunction)env_open_db, METH_VARARGS|METH_KEYWORDS},
     {"path", (PyCFunction)env_path, METH_NOARGS},
     {"stat", (PyCFunction)env_stat, METH_NOARGS},
-    {"sync", (PyCFunction)env_sync, METH_OLDARGS},
+    {"sync", (PyCFunction)env_sync, METH_VARARGS},
+    {"get", (PyCFunction)env_get, METH_VARARGS|METH_KEYWORDS},
+    {"gets", (PyCFunction)env_gets, METH_VARARGS|METH_KEYWORDS},
+    {"put", (PyCFunction)env_put, METH_VARARGS|METH_KEYWORDS},
+    {"puts", (PyCFunction)env_puts, METH_VARARGS|METH_KEYWORDS},
+    {"delete", (PyCFunction)env_delete, METH_VARARGS|METH_KEYWORDS},
+    {"deletes", (PyCFunction)env_deletes, METH_VARARGS|METH_KEYWORDS},
+    {"cursor", (PyCFunction)env_cursor, METH_VARARGS|METH_KEYWORDS},
     {NULL, NULL}
 };
 
@@ -1065,11 +1647,11 @@ cursor_clear(CursorObject *self)
         self->valid = 0;
     }
     if(self->key_buf) {
-        self->key_buf->b_size = 0;
+        SET_BUFFER(self->key_buf, "", 0);
         Py_CLEAR(self->key_buf);
     }
     if(self->val_buf) {
-        self->val_buf->b_size = 0;
+        SET_BUFFER(self->val_buf, "", 0);
         Py_CLEAR(self->val_buf);
     }
     if(self->item_tup) {
@@ -1087,37 +1669,6 @@ cursor_dealloc(CursorObject *self)
     PyObject_Del(self);
 }
 
-static PyObject *
-make_cursor(DbObject *db, TransObject *trans)
-{
-    if(! trans->valid) {
-        return err_invalid();
-    }
-    if(! db) {
-        db = trans->env->main_db;
-    }
-
-    CursorObject *self = PyObject_New(CursorObject, &PyCursor_Type);
-    int rc = mdb_cursor_open(trans->txn, db->dbi, &self->curs);
-    if(rc) {
-        PyObject_Del(self);
-        return err_set("mdb_cursor_open", rc);
-    }
-
-    DEBUG("sizeof cursor = %d", (int) sizeof *self)
-    OBJECT_INIT(self)
-    LINK_CHILD(trans, self)
-    self->positioned = 0;
-    self->key_buf = NULL;
-    self->val_buf = NULL;
-    self->key.mv_size = 0;
-    self->val.mv_size = 0;
-    self->item_tup = NULL;
-    self->trans = trans;
-    Py_INCREF(self->trans);
-    return (PyObject *) self;
-}
-
 static PyObject *
 cursor_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
@@ -1497,6 +2048,57 @@ cursor_iterprev(CursorObject *self, PyObject *args, PyObject *kwargs)
     return iter_from_args(self, args, kwargs, MDB_LAST, MDB_PREV);
 }
 
+static PyObject *
+cursor_iter_from(CursorObject *self, PyObject *args)
+{
+    struct cursor_iter_from {
+        MDB_val key;
+        int reverse;
+    } arg = {{0, 0}, 0};
+
+    static const struct argspec argspec[] = {
+        {ARG_BUF, KEY_S, OFFSET(cursor_iter_from, key)},
+        {ARG_BOOL, REVERSE_S, OFFSET(cursor_iter_from, reverse)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, NULL, &arg)) {
+        return NULL;
+    }
+
+    int rc;
+    if((! arg.key.mv_size) && (! arg.reverse)) {
+        rc = _cursor_get_c(self, MDB_FIRST);
+    } else {
+        self->key = arg.key;
+        rc = _cursor_get_c(self, MDB_SET_RANGE);
+    }
+
+    if(rc) {
+        return NULL;
+    }
+
+    enum MDB_cursor_op op = MDB_NEXT;
+    if(arg.reverse) {
+        op = MDB_PREV;
+        if(! self->positioned) {
+            if(_cursor_get_c(self, MDB_LAST)) {
+                return NULL;
+            }
+        }
+    }
+
+    DEBUG("positioned? %d", self->positioned)
+    IterObject *iter = PyObject_New(IterObject, &PyIterator_Type);
+    if(iter) {
+        iter->val_func = (void *)cursor_item;
+        iter->curs = self;
+        Py_INCREF(self);
+        iter->started = 0;
+        iter->op = op;
+    }
+    return (PyObject *) iter;
+}
+
 static struct PyMethodDef cursor_methods[] = {
     {"count", (PyCFunction)cursor_count, METH_NOARGS},
     {"delete", (PyCFunction)cursor_delete, METH_NOARGS},
@@ -1513,6 +2115,7 @@ static struct PyMethodDef cursor_methods[] = {
     {"set_key", (PyCFunction)cursor_set_key, METH_O},
     {"set_range", (PyCFunction)cursor_set_range, METH_O},
     {"value", (PyCFunction)cursor_value, METH_NOARGS},
+    {"_iter_from", (PyCFunction)cursor_iter_from, METH_VARARGS},
     {NULL, NULL}
 };
 
@@ -1521,7 +2124,7 @@ PyTypeObject PyCursor_Type = {
     .tp_basicsize = sizeof(CursorObject),
     .tp_dealloc = (destructor) cursor_dealloc,
     .tp_clear = (inquiry) cursor_clear,
-    .tp_flags = Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_ITER,
+    .tp_flags = Py_TPFLAGS_DEFAULT,
     .tp_iter = (getiterfunc)cursor_iter,
     .tp_methods = cursor_methods,
     .tp_name = "Cursor",
@@ -1529,10 +2132,9 @@ PyTypeObject PyCursor_Type = {
 };
 
 
-
-/// ------------------------
-// iterator
-// -------------
+// ---------
+// Iterators
+// ---------
 
 static void
 iter_dealloc(IterObject *self)
@@ -1581,7 +2183,7 @@ PyTypeObject PyIterator_Type = {
     PyObject_HEAD_INIT(0)
     .tp_basicsize = sizeof(IterObject),
     .tp_dealloc = (destructor) iter_dealloc,
-    .tp_flags = Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_ITER,
+    .tp_flags = Py_TPFLAGS_DEFAULT,
     .tp_iter = (getiterfunc)iter_iter,
     .tp_iternext = (iternextfunc)iter_next,
     .tp_methods = iter_methods,
@@ -1589,12 +2191,9 @@ PyTypeObject PyIterator_Type = {
 };
 
 
-/// ------------------------
-//transactions
-//////////////////////
-//
-//
-//
+// ------------
+// Transactions
+// ------------
 
 static int
 trans_clear(TransObject *self)
@@ -1706,33 +2305,8 @@ trans_cursor(TransObject *self, PyObject *args, PyObject *kwds)
 static PyObject *
 trans_delete(TransObject *self, PyObject *args, PyObject *kwds)
 {
-    struct trans_delete {
-        MDB_val key;
-        MDB_val val;
-        DbObject *db;
-    } arg = {{0, 0}, {0, 0}, NULL};
-
-    static const struct argspec argspec[] = {
-        {ARG_BUF, KEY_S, OFFSET(trans_delete, key)},
-        {ARG_BUF, VALUE_S, OFFSET(trans_delete, val)},
-        {ARG_DB, DB_S, OFFSET(trans_delete, db)}
-    };
-
-    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
-        return NULL;
-    }
-    if(! arg.db) {
-        arg.db = self->env->main_db;
-    }
-    MDB_val *val_ptr = arg.val.mv_size ? &arg.val : NULL;
-    int rc = mdb_del(self->txn, arg.db->dbi, &arg.key, val_ptr);
-    if(rc) {
-        if(rc == MDB_NOTFOUND) {
-             Py_RETURN_FALSE;
-        }
-        return err_set("mdb_del", rc);
-    }
-    Py_RETURN_TRUE;
+    return generic_delete(self->valid, self->txn, self->env->main_db,
+                          args, kwds);
 }
 
 
@@ -1766,91 +2340,15 @@ trans_drop(TransObject *self, PyObject *args, PyObject *kwds)
 static PyObject *
 trans_get(TransObject *self, PyObject *args, PyObject *kwds)
 {
-    struct trans_get {
-        MDB_val key;
-        PyObject *default_;
-        DbObject *db;
-    } arg = {{0, 0}, Py_None, self->env->main_db};
-
-    static const struct argspec argspec[] = {
-        {ARG_BUF, KEY_S, OFFSET(trans_get, key)},
-        {ARG_OBJ, DEFAULT_S, OFFSET(trans_get, default_)},
-        {ARG_DB, DB_S, OFFSET(trans_get, db)}
-    };
-
-    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
-        return NULL;
-    }
-
-    if(! arg.key.mv_data) {
-        return type_error("key must be given.");
-    }
-
-    MDB_val val;
-    int rc = mdb_get(self->txn, arg.db->dbi, &arg.key, &val);
-    if(rc) {
-        if(rc == MDB_NOTFOUND) {
-            Py_INCREF(arg.default_);
-            return arg.default_;
-        }
-        return err_set("mdb_get", rc);
-    }
-    if(self->buffers) {
-        return buffer_from_val(&self->key_buf, &val);
-    }
-    return string_from_val(&val);
+    return generic_get(self->valid, self->txn, self->env->main_db,
+                       self->buffers, &self->key_buf, args, kwds);
 }
 
 static PyObject *
 trans_put(TransObject *self, PyObject *args, PyObject *kwds)
 {
-    struct trans_put {
-        MDB_val key;
-        MDB_val value;
-        int dupdata;
-        int overwrite;
-        int append;
-        DbObject *db;
-    } arg = {{0, 0}, {0, 0}, 0, 1, 0, self->env->main_db};
-
-    static const struct argspec argspec[] = {
-        {ARG_BUF, KEY_S, OFFSET(trans_put, key)},
-        {ARG_BUF, VALUE_S, OFFSET(trans_put, value)},
-        {ARG_BOOL, DUPDATA_S, OFFSET(trans_put, dupdata)},
-        {ARG_BOOL, OVERWRITE_S, OFFSET(trans_put, overwrite)},
-        {ARG_BOOL, APPEND_S, OFFSET(trans_put, append)},
-        {ARG_DB, DB_S, OFFSET(trans_put, db)}
-    };
-
-    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
-        return NULL;
-    }
-
-    int flags = 0;
-    if(! arg.dupdata) {
-        flags |= MDB_NODUPDATA;
-    }
-    if(! arg.overwrite) {
-        flags |= MDB_NOOVERWRITE;
-    }
-    if(arg.append) {
-        flags |= MDB_APPEND;
-    }
-
-    DEBUG("inserting '%.*s' (%d) -> '%.*s' (%d)",
-        (int)arg.key.mv_size, (char *)arg.key.mv_data,
-        (int)arg.key.mv_size,
-        (int)arg.value.mv_size, (char *)arg.value.mv_data,
-        (int)arg.value.mv_size)
-
-    int rc = mdb_put(self->txn, (arg.db)->dbi, &arg.key, &arg.value, flags);
-    if(rc) {
-        if(rc == MDB_KEYEXIST) {
-            Py_RETURN_FALSE;
-        }
-        return err_set("mdb_put", rc);
-    }
-    Py_RETURN_TRUE;
+    return generic_put(self->valid, self->txn, self->env->main_db,
+                       args, kwds);
 }
 
 static PyObject *trans_enter(TransObject *self)
@@ -1906,14 +2404,34 @@ static int add_type(PyObject *mod, PyTypeObject *type)
         return -1;
     }
     return PyObject_SetAttrString(mod, type->tp_name, (PyObject *)type);
- }
+}
+
+
+#if PY_MAJOR_VERSION >= 3
+static struct PyModuleDef moduledef = {
+    PyModuleDef_HEAD_INIT,
+    "cpython",
+    NULL,
+    0,
+    NULL,
+    NULL,
+    NULL,
+    NULL,
+    NULL
+};
+#endif
+
 
 PyMODINIT_FUNC
-initcpython(void)
+MODINIT_NAME(void)
 {
+#if PY_MAJOR_VERSION >= 3
+    PyObject *mod = PyModule_Create(&moduledef);
+#else
     PyObject *mod = Py_InitModule3("cpython", NULL, "");
+#endif
     if(! mod) {
-        return;
+        MOD_RETURN(NULL);
     }
 
     static PyTypeObject *types[] = {
@@ -1927,41 +2445,42 @@ initcpython(void)
     int i;
     for(i = 0; types[i]; i++) {
         if(add_type(mod, types[i])) {
-            return;
+            MOD_RETURN(NULL);
         }
     }
 
     string_tbl = malloc(sizeof(PyObject *) * STRING_ID_COUNT);
     if(! string_tbl) {
-        return;
+        MOD_RETURN(NULL);
     }
 
     const char *cur = strings;
     for(i = 0; i < STRING_ID_COUNT; i++) {
-        if(! ((string_tbl[i] = PyString_InternFromString(cur)))) {
-            return;
+        if(! ((string_tbl[i] = PyUnicode_InternFromString(cur)))) {
+            MOD_RETURN(NULL);
         }
         cur += strlen(cur) + 1;
     }
 
     if(! ((py_zero = PyLong_FromSize_t(0)))) {
-        return;
+        MOD_RETURN(NULL);
     }
     if(! ((py_int_max = PyLong_FromSize_t(INT_MAX)))) {
-        return;
+        MOD_RETURN(NULL);
     }
     if(! ((py_size_max = PyLong_FromSize_t(SIZE_MAX)))) {
-        return;
+        MOD_RETURN(NULL);
     }
 
     Error = PyErr_NewException("lmdb.Error", NULL, NULL);
     if(! Error) {
-        return;
+        MOD_RETURN(NULL);
     }
     if(PyObject_SetAttrString(mod, "Error", Error)) {
-        return;
+        MOD_RETURN(NULL);
     }
     if(PyObject_SetAttrString(mod, "open", (PyObject *)&PyEnvironment_Type)) {
-        return;
+        MOD_RETURN(NULL);
     }
+    MOD_RETURN(mod);
 }
